#!/bin/sh /usr/share/centrifydc/perl/run

# Copyright (C) 2009-2020 Centrify Corporation. All rights reserved.
#

#

use strict;
use lib '/usr/share/centrifydc/perl';
use CentrifyDC::Include;
use CentrifyDC::GP::Args;
use CentrifyDC::GP::General qw(:debug);
use CentrifyDC::GP::Lock;
use CentrifyDC::GP::Registry;
use CentrifyDC::GP::General qw(RunCmd GetTempDirPath CreateTempFile);
use CentrifyDC::GP::CimsPath qw($CIMS_USR $CIMS_INSTALL);
use CentrifyDC::GP::GPIsolation qw(GetRegKey);

use Date::Parse;
use File::Glob qw(:glob csh_glob);
use File::Basename;

sub IsCAcertsSupported(); # Check if ca-certificates package is supported

my $TEMP_DIR = GetTempDirPath(0);
defined($TEMP_DIR) or FATAL_OUT();

my $openssl = "$CIMS_INSTALL/bin/openssl";
my $nssdbstore = "/etc/pki/nssdb";
my $get_crl_pl = "$CIMS_INSTALL/sbin/get_crl.pl";
my $temp_derform_crl = "$TEMP_DIR/cdc-derformcrl";
my $c_rehash = "$CIMS_INSTALL/bin/c_rehash";
# c_rehash needs this environment variable for openssl binary
$ENV{OPENSSL} = $openssl;

my $certstore = CentrifyDC::Config::GetCertStore(undef);
my $openssl_certstore = "/etc/pam_pkcs11/cacerts";
my $openssl_crlstore = "/etc/pam_pkcs11/crls";
my $openssl_sys_certstore = "/etc/pki/tls/certs";
my $trusted_ca_store = "/etc/pki/ca-trust/source/anchors";
my $update_ca_trust = "/usr/bin/update-ca-trust";

my $hash;
my $ret;
my $rc;
my $out;

my $args = CentrifyDC::GP::Args->new('machine');
CentrifyDC::GP::Registry::Load(undef);
if ($args->isMap())
{
    # first, make sure we have all the dirs
    my @dirs = (${certstore}, ${openssl_certstore}, ${openssl_crlstore});
    foreach my $dir (@dirs)
    {
        RunCmd(["mkdir", "-p", "-m", "755", $dir]);
    }

    # iterate through all certs and put them in openssl and nssdb location
    my @certs = csh_glob("${certstore}/*.cert");
    foreach my $cert_fullpath (@certs)
    {
        my $cert_file = basename $cert_fullpath;
        # getting rid of ".cert" part of filename
        my $cert_filename = substr $cert_file, 0, -5;

        # get hash
        ($rc, $hash) = RunCmd([$openssl, "x509", "-in", $cert_fullpath, "-inform", "PEM", "-hash", "-noout"]);
        # openssl returns 0 on success, if we get something wrong then don't link it and continue
        if ($rc != 0)
        {
            DEBUG_OUT("skipping this cert because it failed to convert: [$cert_fullpath]");
            next;
        }
        chomp $hash;

        if ($cert_filename =~ /^trust_.*/)
        {
            # import cert into system NSSDB
            RunCmd(["certutil", "-A", "-d", $nssdbstore, "-n", $cert_filename, "-t", "CT,C,C", "-i", $cert_fullpath]);

            # symlink to openssl certstore
            my $symlink_target_cert = "${openssl_certstore}/${cert_filename}.cert";
            my $retval;
            $retval = symlink "${cert_fullpath}", "${symlink_target_cert}";
            if ($retval == 0 && !(-e $symlink_target_cert))
            {
                DEBUG_OUT("symlink failed for [${cert_fullpath}]");
            }

            # let's make sure each hash file is unique...
            my @candidates = <${openssl_certstore}/${hash}.*>;
            my $hash_already_exists = 0;
            foreach my $candidate (@candidates)
            {
                next unless -l $candidate;
                if (readlink($candidate) eq "${cert_fullpath}")
                {
                    $hash_already_exists = 1;
                    last;
                }
            }
            unless ($hash_already_exists)
            {
                my $count = @candidates;
                my $symlink_target_hash = "${openssl_certstore}/${hash}.${count}";
                $retval = symlink "${cert_fullpath}", "${symlink_target_hash}";
                if ($retval == 0 && !(-e $symlink_target_hash)) 
                {
                    DEBUG_OUT("symlink failed for [${cert_fullpath}]");
                }
            }

            # get the crl for this cert
            RunCmd([$get_crl_pl, $cert_fullpath, "", "1"]);
        }
    }
    # iterate through all smartcard certificates and download corresponding crls
    my @sccerts = csh_glob("${certstore}/*.pem");
    foreach my $sccert_fullpath (@sccerts)
    {
        RunCmd([$get_crl_pl, $sccert_fullpath, "", "1"]);
    }
    # iterate through all crls and put them in openssl and nssdb location
    my @crls = csh_glob("${certstore}/*.crl");
    foreach my $crl_fullpath (@crls)
    {
        my $crl_file = basename $crl_fullpath;
        # getting rid of ".crl" part of filename
        my $crl_filename = substr $crl_file, 0, -4;

        # get hash
        ($rc, $hash) = RunCmd([$openssl, "crl", "-in", $crl_fullpath, "-inform", "PEM", "-hash", "-noout"]);
        # openssl returns 0 on success, if we get something wrong then don't link it and continue
        if ($rc != 0)
        {
            DEBUG_OUT("skipping this crl because it failed to convert: [$crl_fullpath]");
            next;
        }
        chomp $hash;

        if ($crl_filename =~ /^trust_.*/)
        {
            # import crl into system NSSDB
            my ($fh, $tmpfile) = CreateTempFile($temp_derform_crl);
            if (! defined($fh) or ! defined($tmpfile))
            {
                DEBUG_OUT("Cannot create temp file");
            }
            else
            {
                close($fh);
                RunCmd([$openssl, "crl", "-in", $crl_fullpath, "-outform", "DER", "-out", $tmpfile]);
                RunCmd(["crlutil", "-I", "-d", $nssdbstore, "-i", $tmpfile]);
                unlink($tmpfile);
            }

            # symlink to openssl certstore
            my $symlink_target_crl = "${openssl_crlstore}/${crl_filename}.crl";
            my $retval;
            $retval = symlink "${crl_fullpath}", "${symlink_target_crl}";
            if ($retval == 0 && !(-e $symlink_target_crl))
            {
                DEBUG_OUT("symlink failed for [${crl_fullpath}]");
            }

            # let's make sure each hash file is unique...
            my @candidates = <${openssl_crlstore}/${hash}.*>;
            my $hash_already_exists = 0;
            foreach my $candidate (@candidates)
            {
                next unless -l $candidate;
                if (readlink($candidate) eq "${crl_fullpath}")
                {
                    $hash_already_exists = 1;
                    last;
                }
            }
            unless ($hash_already_exists)
            {
                my $count = @candidates;
                my $symlink_target_hash = "${openssl_crlstore}/${hash}.r${count}";
                $retval = symlink "${crl_fullpath}", "${symlink_target_hash}";
                if ($retval == 0 && !(-e $symlink_target_hash)) 
                {
                    DEBUG_OUT("symlink failed for [${crl_fullpath}]");
                }
            }
        }
    }

    my @cacerts = csh_glob("${certstore}/trust_*.cert");

    if (IsCAcertsSupported()) 
    {
        DEBUG_OUT("The update-ca-trust is supported");

        # This is mainly for RHEL 6 and newer versions where ca-certificates
        # package is available.
        # Copy all the trusted CA certs to /etc/pki/ca-trust/source/anchors/.
        # our RunCmd facility uses single quote, so no expansion for * enclosed.
        # so iterate through the files. Then call update-ca-trust.

        foreach my $cacert (@cacerts)
        {
            RunCmd(["cp", $cacert, $trusted_ca_store]);
        }
        my ($ret, $out) = RunCmd([$update_ca_trust, "enable"]);
        if ($ret == 0)
        {
            if ( $out =~ m/P15Applet_ReadBinary/ )
            {
                # Ignore error message about "undefine symbol: P15Applet_ReadBinary"
                # as our coolkey library does not know P15Applet_ReadBinary when
                # p11-kit parse key usage from X509 certificate
                # This error does not affect the functionality of update-ca-trust
                # Please refer CC-56176 for detailed information
            }
            else
            {
                ERROR_OUT($out);
            }
        }
        my ($ret, $out) = RunCmd([$update_ca_trust, "extract"]);
        if ($ret == 0)
        {
            if ( $out =~ m/P15Applet_ReadBinary/ )
            {
                # Ignore error message about "undefine symbol: P15Applet_ReadBinary"
                # as our coolkey library does not know P15Applet_ReadBinary when
                # p11-kit parse key usage from X509 certificate
                # This error does not affect the functionality of update-ca-trust
                # Please refer CC-56176 for detailed information
            }
            else
            {
                ERROR_OUT($out);
            }
        }
    }
    else
    {
        DEBUG_OUT("The update-ca-trust is NOT supported");
        # This is mainly for RHEL 5 and older versions where ca-certificates package is
        # not available. 
        # Copy all the trusted CA certs to /etc/pki/tls/certs/.
        # our RunCmd facility uses single quote, so no expansion for * enclosed.
        # so iterate through the files. Then call c_rehash. This is for
        # non-curl openssl aware applications.
        # all OpenSSL based applications recognize the convention of hasing
        # names except for those directly using CA truststore location such as
        # cURL.
        # for cURL, we need to append the GPO pushed CA root cert to the system
        # default bundle file at /etc/pki/tls/certs/ca-bundle.crt so that when
        # running curl without --cacert would work as well.
        # check if it was already appended before appending the ca cert to the default truststore.
        if (-f "${openssl_sys_certstore}/ca-bundle.crt")
        {
            my $anychange = 0;
            my $bundle = do {
                open (my $fh2, "${openssl_sys_certstore}/ca-bundle.crt") or die "Cannot open file: ${openssl_sys_certstore}/ca-bundle.crt";
                local $/ = undef;
                <$fh2>;
            };
            # Remove newlines and special chars. Normalization.
            $bundle =~ tr /\n//d;
            $bundle =~ s/\W//g;

            foreach my $cacert (@cacerts)
            {
                my $gpocert = do {
                    open (my $fh1, $cacert) or die "Cannot open file: $cacert";
                    local $/ = undef;
                    <$fh1>;
                };
                # Remove newlines and special chars. Normalization.
                $gpocert =~ tr /\n//d;
                $gpocert =~ s/\W//g;

                if ($bundle =~ /$gpocert/) {
                    # Already exists, do nothing. 
                }
                else
                {
                    $anychange = 1;
                    my $appendcmd = "$openssl x509 -in $cacert >> ${openssl_sys_certstore}/ca-bundle.crt";
                    RunCmd([$appendcmd]);
                    RunCmd(["cp", $cacert, $openssl_sys_certstore]);
                }
            }
            if ($anychange)
            {
                RunCmd([$c_rehash, $openssl_sys_certstore]);
            }
        }
        else
        {
            DEBUG_OUT("The default system certstore: ${openssl_sys_certstore}/ca-bundle.crt does not exist!");
        }
    }
    # cleanup symlinks
    my @certItems = csh_glob("${openssl_certstore}/*");
    my @crlItems = csh_glob("${openssl_crlstore}/*");
    my @deleteItems;
    foreach my $item (@certItems, @crlItems)
    {
        # if the file does not exist but the symlink exists, clean it up
        if ( ! -e $item ) 
        {
            DEBUG_OUT("Cleaning up symlink $item since it points to an empty file");
            push(@deleteItems, $item);
        }
    }
    unlink(@deleteItems);
}
else
{
    # remove symlinks
    my @certItems = csh_glob("${openssl_certstore}/*");
    my @crlItems = csh_glob("${openssl_crlstore}/*");
    my @deleteItems;

    foreach my $item (@certItems)
    {
        if ($item !~ m/^${openssl_certstore}\/auto_.*/)
        {
            push(@deleteItems, $item);
            my $cert_file = basename $item;
            my $cert_filename = substr $cert_file, 0, -5;
            RunCmd(["certutil", "-D", "-d", $nssdbstore, "-n", $cert_filename]);
        }
    }

    foreach my $item (@crlItems)
    {
        if ($item !~ m/^${openssl_crlstore}\/auto_.*/)
        {
            push(@deleteItems, $item);
            my $crl_file = basename $item;
            my $crl_filename = substr $crl_file, 0, -4;
            RunCmd(["crlutil", "-D", "-d", $nssdbstore, "-n", $crl_filename]);
        }
    }

    unlink(@deleteItems);
}

# 
# Check if the current OS supports ca-certificates package.
# We use the update-ca-trust tool to update CA root if possible.
# RHEL 6 and newer support this tool.
#
#   return: 1       - supported
#           0       - not supported
#           undef   - failed
#
#
sub IsCAcertsSupported()
{
    my $supported = 0;

    if (-f $update_ca_trust && -x $update_ca_trust)
    {
        $supported = 1;
    }

    return $supported;
}
