#!/bin/sh /usr/share/centrifydc/perl/run

##############################################################################
#
# Copyright (C) 2014-2020 Centrify Corporation. All rights reserved.
#
# RHEL-specific machine mapper script to enable/disable smartcard support
#
#  The script calls sctool to enable/disable smartcard support.
#
#  We can't get current system setting whether smartcard support is enabled
#  or not.
#
#  Map:     enable or disable smartcard support based on gp setting
#             Not configured:  disable
#             Enabled:         enable
#             Disabled:        disable
#
#  Unmap:   disable smartcard support
#
# Parameters: <map|unmap> mode
#   map|unmap   action to take
#   mode        mode (not used)
#
# Exit value:
#   0   Normal
#   1   Error
#   2   usage
#
##############################################################################

use strict;
use lib '/usr/share/centrifydc/perl';
use CentrifyDC::Include;
use CentrifyDC::GP::Args;
use CentrifyDC::GP::Registry;
use CentrifyDC::GP::General qw(:debug RunCmd GetTempDirPath IsEqual);
use CentrifyDC::GP::Mapper;
use CentrifyDC::GP::RegHelper;


# >>> DATA >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

my $TEMP_DIR = GetTempDirPath(0);
defined($TEMP_DIR) or FATAL_OUT();


my $SCTOOL = '/usr/bin/sctool';

my $REGKEY = 'software/policies/centrify/unixsettings/security/rhelsmartcardenabled';
my $REGKEY_SMARTCARDENABLED = "RHELSmartCardEnabled";
my $REGKEY_PKCS11MODULE = "RHELSmartCardPKCS11Module";

my $LIB = `uname -m`;
chomp($LIB);
if ($LIB eq 'x86_64')
{
    $LIB = 'lib64';
}
else
{
    $LIB = 'lib';
}

my $PKCS11_MODULE_PATH = '/usr/' . $LIB . '/pkcs11/libcentrifypkcs11.so';

# standard
sub HandleSctoolReturnCode($);
sub ToggleSmartCard($);
sub DisableSmartCard();
sub Map();
sub UnMap();

# >>> SUB >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>


# Handle sctool return codes
#
#   return: 1       - successful
#           undef   - failed
#
sub HandleSctoolReturnCode($)
{
    my $ret = $_[0];

    if (defined($ret))
    {
        # sctool will return SCTOOL_OK (0) when success
        # so we need to convert return value here
        if ($ret eq '0')
        {
            $ret = 1;
        }
        else
        {
            my $msg = "sctool returned [$ret] : ";

            # SCTOOL_ERR_RPM_DEP (-1)
            if ($ret eq '255')
            {
        	$msg .= "RPM dependency check failed.";
            }
            # SCTOOL_ERR_LOCK_FILE_{CREATE,LOCK} (-4, -5)
            elsif ($ret eq '252' or $ret eq '251')
            {
                $msg .= "Failed to access lock file.";
            }
            # SCTOOL_ERR_PAM_FILE (-6)
            elsif ($ret eq '250')
            {
                $msg .= "Failed to access PAM file.";
            }
            # SCTOOL_ERR_TEMP_FILE (-7)
            elsif ($ret eq '249')
            {
                $msg .= "Failed to access temp file.";
            }
            # SCTOOL_ERR_GENERAL_EXCEPTION (-8)
            elsif ($ret eq '248')
            {
                $msg .= "Generic error.";
            }
            else
            {
                $msg .= "Unknown error.";
            }

            $msg .= " Please consult verbose log (i.e. TRACE) for more information.";

            INFO_OUT($msg);

            $ret = undef;
        }
    }
    return $ret;
}

sub ToggleSmartCard($)
{
    my $action = $_[0];
    my $cmd;
    my $ret = 1;

    my $reg = CentrifyDC::GP::RegHelper->new($action, 'machine', $REGKEY, undef, undef);
    $reg->load();

    my $sys_setting = { };
    
    my $sc_status = RunCmd([$SCTOOL, "--status"]);

    # Convert the return code into GP value
    if ($sc_status eq '0' or $sc_status eq '1')
    {
        $sys_setting->{$REGKEY_SMARTCARDENABLED} = "YES";
    }
    else
    {
        $sys_setting->{$REGKEY_SMARTCARDENABLED} = "NO";
    }
    
    # A better idea is to make sctool report which pkcs #11 module we are
    # using, as the underlying implementation would be changed in the future
    $sys_setting->{$REGKEY_PKCS11MODULE} = readlink($PKCS11_MODULE_PATH);

    # Since we support $LIB variable, we have to expand it before we can
    # compare it againsh previous/system setting.
    my $reg_cur = $reg->get('current');
    $reg_cur->{$REGKEY_PKCS11MODULE} =~ s/(\$\w+)/$1/eeg;
    my $reg_pre = $reg->get('previous');
    $reg_pre->{$REGKEY_PKCS11MODULE} =~ s/(\$\w+)/$1/eeg;

    $reg->set('system', $sys_setting) or return undef;

    my $group = $reg->getGroupToApply();
    
    if ($group)
    {
        my $data = $reg->get($group);
        my $lock_file = "$TEMP_DIR/smartcard.lock";

        if ($data->{$REGKEY_SMARTCARDENABLED} eq "YES")
        {
            $ret = RunCmd([$SCTOOL, "--disable"], {'lockfile' => $lock_file});
            $ret = RunCmd([$SCTOOL, "--enable"], {'lockfile' => $lock_file});
        }
        else
        {
            $ret = RunCmd([$SCTOOL, "--disable"], {'lockfile' => $lock_file});
        }

        if (-x "/usr/sbin/gdm-safe-restart")
        {
            RunCmd(["/usr/sbin/gdm-safe-restart"]);
        }

        $ret = HandleSctoolReturnCode($ret);
    }

    return $ret;
}

# Force disable smart card
#
#   return: 1       - successful
#           undef   - failed
#
sub DisableSmartCard()
{
    my $ret = undef;

    # disable smart card
    my $cmd = [$SCTOOL, "--disable"];
    my $lock_file = "$TEMP_DIR/smartcard.lock";
    $ret = RunCmd($cmd, {'lockfile' => $lock_file});
    $ret = HandleSctoolReturnCode($ret);

    return $ret;
}

#
# map
#
#   return: 1       - successful
#           undef   - failed
#
sub Map()
{
    return ToggleSmartCard('map');
}

#
# unmap
#
#   return: 1       - successful
#           undef   - failed
#
sub UnMap()
{
    # TODO:
    # Need to revisit why we cannot use ToggleSmartCard('unmap') to pass BAT 
    # test as described in bug 34166. This is not appropriate because we should 
    # restore orginial system settings during unmap, and that does not always 
    # mean sctool is disabled (customer may manually enable sctool before). 
    #
    # Note that mac_mapper_smartcard.pl still calls ToggleSmartCard('unmap') 
    # instead of force disable sctool here.

    # make sure sctool is disabled before leaving a domain
    my $ret = DisableSmartCard();

    return $ret;
}


# >>> DATA >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

#
# centrifydc.conf settings
#
my $file = {
    'hierarchy_separator' => '.',
    'list_expr' => ', *| +',
    'list_separator' => ', ',
    'named_list_separator' => ',',
    'match_expr' => [
      '/^\s*([^\s:=]+)[:=]\s*(.*)/',
    ],
    'comment_markers' => [ '#', ],
    path => '/etc/centrifydc/centrifydc.conf',
    lock => '/etc/centrifydc/centrifydc.conf.lock',
    post_action => [ 'DO_ADRELOAD', 'DO_DSRELOAD', 'DO_SARESTART' ],
    value_map => {
      'rhel.smartcard.login.force' => {
        reg_class   => 'machine',
        reg_key     => 'software/policies/centrify/unixsettings/security/rhelsmartcardloginforce',
        reg_value   => 'RHELSmartCardLoginForce',
        reg_type    => ['REG_SZ'],
        value_type  => 'named',
      },
      'rhel.smartcard.login.force.exceptiongroup' => {
        reg_class   => 'machine',
        reg_key     => 'software/policies/centrify/unixsettings/security/rhelsmartcardloginforceexceptiongroup',
        reg_value   => 'RHELSmartCardLoginForceExceptionGroup',
        reg_type    => ['REG_SZ'],
        value_type  => 'named',
        named_list  => 1,
      },
      'rhel.smartcard.pkcs11.module' => {
        reg_class   => 'machine',
        reg_key     => 'software/policies/centrify/unixsettings/security/rhelsmartcardenabled',
        reg_value   => 'RHELSmartCardPKCS11Module',
        reg_type    => ['REG_SZ'],
        value_type  => 'named',
      },
    },
    'write_data' => '$value: $data\n',
};

# >>> MAIN >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

my $args = CentrifyDC::GP::Args->new('machine');

CentrifyDC::GP::Registry::Load(undef);

my $ret = 0;

if ($args->isMap())
{
    CentrifyDC::GP::Mapper::Map($file, $args->user());
    $ret = Map();
}
else
{
    CentrifyDC::GP::Mapper::UnMap($file, $args->user());
    $ret = UnMap();
}

$ret or FATAL_OUT();

