#!/bin/sh /usr/share/centrifydc/perl/run

##############################################################################
#
# Copyright (C) 2004-2020 Centrify Corporation. All rights reserved.
#
# User mapper script to apply Gnome 3 settings.
#
#
#   The mapper script gets all Gnome 3 GP settings (current/previous/local),
#   compare with system's Gnome 3 settings to determine what GP settings can
#   be applied to current system, then determine settings that need to change.
#
#   If user has existing Gnome sessions, then connect to all these sessions
#   and change settings in all sessions. Settings will take effect immediately.
#
#   If user has no Gnome session, create a DBUS session, change settings
#   in this session, then kill the session. Settings will take effect
#   immediately after user login.
#
#   To tackle special characters in settings, gsettings commands will be
#   written into a script file to execute.
#
#   The script returns the number of failed gsettings command. If there's
#   error, get system settings again to check which setting is not applied.
#   This is not 100% accurate, as for float number, what you set may be
#   slightly different from what you get, for example setting value to 0.05
#   may end up as 0.050000000000000003.
#
#   The mapper script doesn't handle relocatable settings.
#
#
#   Map:     Apply Gnome 3 Group Policy settings
#   Unmap:   Restore local Gnome 3 settings
#
# Parameters: <map|unmap> <username> <mode>
#   map|unmap   action to take
#   username    username
#   mode        not used
#
# Exit value:
#   0   Normal
#   1   Error
#   2   Usage
#
##############################################################################

use strict;
use lib '/usr/share/centrifydc/perl';
use CentrifyDC::Include;
use POSIX;
use File::Basename qw(basename);

use CentrifyDC::GP::Args;
use CentrifyDC::GP::General qw(:debug IsEmpty IsTraceOn RunCmd ReadFile WriteFile GetTempDirPath);
use CentrifyDC::GP::RegHelper;
use CentrifyDC::GP::Registry;

my $DBUSLAUNCH = '/usr/bin/dbus-launch';
my $DCONF      = '/usr/bin/dconf';
my $GSETTINGS  = '/usr/bin/gsettings';
my $PS         = '/bin/ps';
my $SU         = '/bin/su';

my $REGKEY = "software/policies/centrify/gnome3";
my $REGKEY_EXTRA = "software/policies/centrify/gnome3custom";
my $MAX_DBUS_SESSIONS = 20;
my $GSETTINGS_SCRIPT_FILE = 'Gnome3.sh';

my $g_relocatable_schemas = {}; # ignore relocatable schemas
my $g_schemas = {};             # lowercase schema id -> real schema id

sub IsGnome3();
sub IsRoot();
sub CheckTools();
sub GetRelocatableSchemas($);
sub GetSysSettings($);
sub _PrintSysSettings($);

sub GetCombinedSettings($$$);
sub GetCombinedCustomSettings($$$);
sub GetAllSettings($$$);
sub GetChangedSettings($$);
sub CreateScript($);
sub UpdateSettings($$$);
sub VerifySettings($$);
sub GetAction($);
sub CreateDbusSession($);
sub KillDbusSession($);
sub GetExistingDbusSessions($);
sub DoMap($$);



# >>> SUB >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

#
# Check if Gnome 3 is installed
#
# Run gnome-terminal command to get its version. Only Gnome 3 has valid output.
#
#   return: 1       - yes
#           0       - no
#           undef   - cannot identify
#
sub IsGnome3()
{
    my $GNOME_TERMINAL = '/usr/bin/gnome-terminal';

    if (! -e $GNOME_TERMINAL || ! -x $GNOME_TERMINAL)
    {
        DEBUG_OUT("$GNOME_TERMINAL doesn't exist or is not executable.");
        return 0;
    }

    my $cmd = [$GNOME_TERMINAL, "--version"];
    my ($rc, $output) = RunCmd($cmd);
    if (! defined($rc) or $rc ne '0')
    {
        DEBUG_OUT("Fail to run command [$cmd].");
        return undef;
    }

    if ($output =~ m/GNOME Terminal 3/)
    {
        return 1;
    }
    else
    {
        DEBUG_OUT("Gnome version is not 3.x.");
        return 0;
    }
}


#
# Check if user run this script with root privilege.
#
#   return: 1   - yes
#           0   - no
#
sub IsRoot()
{
    if ($> != 0)
    {
        DEBUG_OUT("Current user is not root.");
        return 0;
    }
    else
    {
        return 1;
    }
}

#
# Check if required tools exists.
#
#   return: 1           - installed
#           undef       - not installed
#
sub CheckTools()
{
    if (! -e $GSETTINGS || ! -x $GSETTINGS)
    {
        DEBUG_OUT("$GSETTINGS doesn't exist or is not executable.");
        return undef;
    }

    if (! -e $DBUSLAUNCH || ! -x $DBUSLAUNCH)
    {
        DEBUG_OUT("$DBUSLAUNCH doesn't exist or is not executable.");
        return undef;
    }

    if (! -e $PS || ! -x $PS)
    {
        DEBUG_OUT("$PS doesn't exist or is not executable.");
        return undef;
    }

    if (! -e $SU || ! -x $SU)
    {
        DEBUG_OUT("$SU doesn't exist or is not executable.");
        return undef;
    }

    return 1;
}

#
# Get relocatable schemas. These settings will be skipped. This is necessary
# on Ubuntu, as gsettings list-recursively also gets relocatable schema.
#
#   $_[0]:  user
#
#   return: 1       - successful
#           undef   - failed
#
sub GetRelocatableSchemas($)
{
    my $user = $_[0];

    defined($user) or return undef;

    my $cmd = [$SU, "-", $user, "-c", "$GSETTINGS list-relocatable-schemas"];
    my ($rc, $output) = RunCmd($cmd);
    if (! defined($rc) or $rc ne '0')
    {
        DEBUG_OUT("Fail to run command [$cmd].");
        return undef;
    }

    my @lines = split(/\n/, $output);

    %$g_relocatable_schemas = map { $_, 1 } @lines;

    return 1;
}

#
# Dump current system Gnome 3 settings into a hash reference.
#
# Use gsettings to dump settings, then parse the output line by line.
# Each line has the format:
#   org.gnome.Vino authentication-methods ['none']
#
#   $_[0]:  user
#
#   return: hash reference  - system settings
#               => {
#                   'org.gnome.system.proxy' => {
#                       'use-same-proxy' => 'true',
#                       'mode'           => 'manual',
#                   },
#           undef           - failed
#
sub GetSysSettings($)
{
    my $user = $_[0];

    defined($user) or return undef;

    my $sys_settings = {};

    my $cmd = [$SU, "-", $user, "-c", "$GSETTINGS list-recursively"];
    my ($rc, $output) = RunCmd($cmd);
    if (! defined($rc) or $rc ne '0')
    {
        DEBUG_OUT("Fail to run command [$cmd].");
        return undef;
    }

    my @lines = split(/\n/, $output);

    foreach my $line (@lines)
    {
        if ($line =~ m/^([^ ]+) ([^ ]+) (.*)/)
        {
            my ($gnome_schema_id, $gnome_key, $gnome_val) = ($1, $2, $3);

            # skip relocatable schemas
            exists($g_relocatable_schemas->{$gnome_schema_id}) and next;

            #
            # some array/hash value has complex format if value is empty, for example
            # clock's alarm setting:
            #   org.gnome.clocks alarms @aa{sv} []
            # these empty values all start with @.
            # if value not empty, then the format will be like
            #   org.gnome.clocks alarms [{'name': <'Alarm'>, 'active': <true>, 'hour': <13>, 'minute': <52>, 'days': <@ai []>}]
            # just remove the string that starts with @ and ends with space
            # character.
            if ($gnome_val =~ m/^@/)
            {
                $gnome_val =~ s/^@[^ ]* //;
            }

            #
            # string values are quoted with ''. remove the quote.
            # '' means empty string value.
            #
            if ($gnome_val =~ m/^'.*'$/)
            {
                $gnome_val =~ s/^'//;
                $gnome_val =~ s/'$//;
            }

            # some settings has uint16/uint32/uint64 prefix, remove the prefix.
            $gnome_val =~ s/^uint[0-9]+ //;

            #
            # some settings may have special characters in it, for example %c.
            # the debug log will be garbled, but actual setting is correct.
            # need to fix log functions later.
            #

            #
            # write settings into hash table. Need to create an entry even if
            # value is undefined, as we need to know if a GP setting is
            # applicable on this platform.
            # 
            $sys_settings->{$gnome_schema_id}{$gnome_key} = $gnome_val;
        }
    }

    # now create a hash table of lowercase schema id -> real schema id
    foreach my $id (keys %$sys_settings)
    {
        $g_schemas->{lc($id)} = $id;
    }

    if (IsTraceOn)
    {
        _PrintSysSettings($sys_settings);
    }

    return $sys_settings;
}

#
# Print current system Gnome 3 settings.
#
#   $_[0]:  hash reference of system settings
#
sub _PrintSysSettings($)
{
    my $sys_settings = $_[0];

    foreach my $gnome_schema_id (keys %$sys_settings)
    {
        DEBUG_OUT("[$gnome_schema_id]");
        my $schemas = $sys_settings->{$gnome_schema_id};
        foreach my $gnome_key (keys %$schemas)
        {
            TRACE_OUT("   $gnome_key:  %s", defined($schemas->{$gnome_key}) ? $schemas->{$gnome_key} : "<no setting>");
        }
    }
}

#
# Get Gnome 3 registry settings and system settings, and put into a hash.
#
#   $_[0]:  action
#   $_[1]:  user
#   $_[2]:  hash reference of system settings
#
#   return: hash reference  - settings
#               => {
#                   'org.gnome.system.proxy' => {
#                       'use-same-proxy' => {
#                           'current'  => 'true',
#                           'previous' => 'false',
#                           'local'    => 'true',
#                           'system'   => 'false',
#                   },
#           undef           - failed
#
sub GetCombinedSettings($$$)
{
    my $action = $_[0];
    my $user   = $_[1];
    my $sys_settings = $_[2];

    defined($user) or return undef;

    # get all sub keys under Gnome 3 GP root key, including current/previous/local
    my @tmp_keys = CentrifyDC::GP::Registry::GetSubKeys($REGKEY, undef, $user);

    # remove duplicated entries
    my %tmp_hash = map { $_, 1 } @tmp_keys;
    my @subkeys = ();

    # remove Gnome 3 GP root key
    foreach my $key (keys %tmp_hash)
    {
        if ($key ne $REGKEY)
        {
            push(@subkeys, $key); 
        }
    }

    my $tmp_reg_settings  = {};
    my $combined_settings = {};
    foreach my $regkey (@subkeys)
    {
        #
        # get gnome schema id from registry key:
        # registry key:
        #   software/policies/centrify/gnome3/org.gnome.system.proxy
        # gnome schema id:
        #   org.gnome.system.proxy
        # 
        my $id = basename($regkey);
        defined($id) or next;

        #
        # if the gnome schema id doesn't exist in system, then registry settings
        # won't work. just skip these settings.
        # please note that gnome schema id can contain upper case letter, but
        # registry key is all lower case, so cannot compare directly.
        #
        if (! exists($g_schemas->{$id}))
        {
            TRACE_OUT("Gnome 3 schema [$id] doesn't exist on current system. Skip.");
            next;
        }

        # now get the real schema id (that may contain uppercase character)
        my $gnome_schema_id = $g_schemas->{$id};

        # first get all gp settings under the registry key
        TRACE_OUT("Search registry key for Gnome 3 settings: [$regkey]");
        $tmp_reg_settings->{$regkey} = CentrifyDC::GP::RegHelper->new($action, 'user', $regkey, undef, undef);
        $tmp_reg_settings->{$regkey} or return undef;
        $tmp_reg_settings->{$regkey}->load();

        foreach my $group (qw(current previous local))
        {
            my $settings = $tmp_reg_settings->{$regkey}->get($group);
            IsEmpty(%$settings) and next;
            foreach my $gnome_key (keys %$settings)
            {
                #
                # even if the gnome schema id exists in system, certain
                # gnome key may not. skip these keys.
                #
                if (! exists($sys_settings->{$gnome_schema_id}{$gnome_key}))
                {
                    TRACE_OUT("Gnome 3 key [$gnome_schema_id:$gnome_key] doesn't exist on current system. Skip.");
                    next;
                }
                $combined_settings->{$gnome_schema_id}{$gnome_key}{$group} = $settings->{$gnome_key};
                # set system setting in the hash
                if (! exists($combined_settings->{$gnome_schema_id}{$gnome_key}{'system'}))
                {
                    $combined_settings->{$gnome_schema_id}{$gnome_key}{'system'} = $sys_settings->{$gnome_schema_id}{$gnome_key};
                }
            }
        }
    }

    return $combined_settings;
}

#
# Get custom Gnome 3 registry settings and system settings, and put into a hash.
#
# Custom Gnome 3 settings are all stored under $REGKEY_EXTRA, registry value
# format is schema:id, for example:
#   org.gnome.system.proxy:use-same-proxy true
#
#   $_[0]:  action
#   $_[1]:  user
#   $_[2]:  hash reference of system settings
#
#   return: hash reference  - settings
#               => {
#                   'org.gnome.system.proxy' => {
#                       'use-same-proxy' => {
#                           'current'  => 'true',
#                           'previous' => 'false',
#                           'local'    => 'true',
#                           'system'   => 'false',
#                   },
#           undef           - failed
#
sub GetCombinedCustomSettings($$$)
{
    my $action = $_[0];
    my $user   = $_[1];
    my $sys_settings = $_[2];

    defined($user) or return undef;

    my $combined_settings = {};

    my $tmp_reg_settings  = {};

    my $regkey = $REGKEY_EXTRA;
    TRACE_OUT("Search registry for custom Gnome 3 settings");
    $tmp_reg_settings->{$regkey} = CentrifyDC::GP::RegHelper->new($action, 'user', $regkey, undef, undef);
    $tmp_reg_settings->{$regkey} or return undef;
    $tmp_reg_settings->{$regkey}->load();

    # there's no local setting
    foreach my $group (qw(current previous))
    {
        my $settings = $tmp_reg_settings->{$regkey}->get($group);
        IsEmpty(%$settings) and next;
        foreach my $str (keys %$settings)
        {
            # schema:key
            if ($str =~ m/([^:]+):([^:]+)/)
            {
                my $schema = lc($1);
                my $key = $2;
                if (! exists($g_schemas->{$schema}))
                {
                    TRACE_OUT("Gnome 3 schema [$schema] doesn't exist on current system. Skip.");
                    next;
                }

                $schema = $g_schemas->{$schema};

                if (! exists($sys_settings->{$schema}{$key}))
                {
                    TRACE_OUT("Gnome 3 key [$schema:$key] doesn't exist on current system. Skip.");
                    next;
                }
                $combined_settings->{$schema}{$key}{$group} = $settings->{$str};
                # set system setting in the hash
                if (! exists($combined_settings->{$schema}{$key}{'system'}))
                {
                    $combined_settings->{$schema}{$key}{'system'} = $sys_settings->{$schema}{$key};
                }
            }
        }
    }

    return $combined_settings;
}

#
# Get all Gnome 3 registry settings and system settings, and put into a hash.
#
# Get settings from both standard Gnome 3 policy and custom Gnome 3 policy,
# then merge them together. When merging, use custom setting only if a gnome
# schema:key only exists in custom setting.
#
#   $_[0]:  action
#   $_[1]:  user
#   $_[2]:  hash reference of system settings
#
#   return: hash reference  - settings
#               => {
#                   'org.gnome.system.proxy' => {
#                       'use-same-proxy' => {
#                           'current'  => 'true',
#                           'previous' => 'false',
#                           'local'    => 'true',
#                           'system'   => 'false',
#                   },
#           undef           - failed
#
sub GetAllSettings($$$)
{
    my $action = $_[0];
    my $user   = $_[1];
    my $sys_settings = $_[2];

    defined($user) or return undef;

    my $combined_settings = GetCombinedSettings($action, $user, $sys_settings);
    my $combined_custom_settings = GetCombinedCustomSettings($action, $user, $sys_settings);

    if (! IsEmpty($combined_custom_settings))
    {
        foreach my $schema (keys %$combined_custom_settings)
        {
            #
            # do a deep copy. if custom setting exists but standard setting
            # doesn't, copy setting from custom setting to standard setting.
            #
            my $tmp_settings = $combined_custom_settings->{$schema};
            if (! exists($combined_settings->{$schema}))
            {
                foreach my $key (keys %$tmp_settings)
                {
                    foreach my $group (qw(current previous local system))
                    {
                        if (exists($tmp_settings->{$key}{$group}))
                        {
                            TRACE_OUT("Add custom setting: $schema:$key  group: $group");
                            $combined_settings->{$schema}{$key}{$group} = $combined_custom_settings->{$schema}{$key}{$group};
                        }
                    }
                }
            }
            else
            {
                foreach my $key (keys %$tmp_settings)
                {
                    #
                    # copy only if a gnome schema:key only exists in custom
                    # setting
                    #
                    if (! exists($combined_settings->{$schema}{$key}{'current'}) and ! exists($combined_settings->{$schema}{$key}{'previous'}))
                    {
                        foreach my $group (qw(current previous system))
                        {
                            if (exists($tmp_settings->{$key}{$group}))
                            {
                                TRACE_OUT("Add custom setting: $schema:$key group: $group");
                                $combined_settings->{$schema}{$key}{$group} = $combined_custom_settings->{$schema}{$key}{$group};
                            }
                        }
                    }
                }
            }
        }
    }

    foreach my $schema (%$combined_settings)
    {
        my $settings = $combined_settings->{$schema};
        foreach my $key (keys %$settings)
        {
            foreach my $group (qw(current previous local system))
            {
                if (exists($settings->{$key}{$group}))
                {
                    TRACE_OUT("combined settings: $schema: $key: $group: $settings->{$key}{$group}");
                }
            }
        }
    }

    return $combined_settings;
}

#
# Get Gnome 3 settings that need to update.
#
#   $_[0]:  combined settings
#   $_[1]:  action
#
#   return: hash reference  - changed settings
#               => {
#                   'org.gnome.system.proxy' => {
#                       'use-same-proxy' => 'true',
#                       'mode'           => 'manual',
#                   },
#
#           undef           - failed
#
sub GetChangedSettings($$)
{
    my $combined_settings = $_[0];
    my $action = $_[1];

    defined($combined_settings) or return undef;
    defined($action) or return undef;

    DEBUG_OUT("Get changed Gnome 3 settings.");

    my $changed_settings = {};
    my $reg_settings = {};

    #
    # now we have a hash of registry+system setting, now create RegHelper
    # instances for them. each Gnome setting has a corresponding RegHelper
    # instance.
    #
    foreach my $gnome_schema_id (keys %$combined_settings)
    {
        my $reg_key = "$REGKEY/$gnome_schema_id";
        my $settings = $combined_settings->{$gnome_schema_id};
        foreach my $gnome_key (keys %$settings)
        {
            $reg_settings->{$gnome_schema_id}{$gnome_key} = CentrifyDC::GP::RegHelper->new($action, 'user', $reg_key, $gnome_key, undef);
            $reg_settings->{$gnome_schema_id}{$gnome_key} or return undef;
            foreach my $group (qw(current previous local system))
            {
                my $data = $settings->{$gnome_key}{$group};
                $reg_settings->{$gnome_schema_id}{$gnome_key}->set($group, $data);
            }
        }
    }

    foreach my $gnome_schema_id (keys %$reg_settings)
    {
        my $settings = $reg_settings->{$gnome_schema_id};
        foreach my $gnome_key (keys %$settings)
        {
            my $group = $settings->{$gnome_key}->getGroupToApply();
            $group or next;
            my $data = $settings->{$gnome_key}->get($group);
            # if no setting, do nothing
            defined($data) or next;
            $changed_settings->{$gnome_schema_id}{$gnome_key} = $data;
            DEBUG_OUT("Need to update: $gnome_schema_id: $gnome_key: [$data]");
        }
    }

    return $changed_settings;
}

#
# Create script that uses gsettings command to change settings.
#
#   $_[0]:  hash reference of changed settings
#
#   return: string  - script path
#           undef   - failed
#
sub CreateScript($)
{
    my $changed_settings = $_[0];

    defined($changed_settings) or return undef;

    # generate script file content
    my $script_str = "#!/bin/sh

i_fail=0
i_success=0

verify_result()
{
";
    $script_str .= '  if [ $? -ne 0 ]; then';
    $script_str .= "\n";
    $script_str .= '    i_fail=`expr $i_fail + 1`';
    $script_str .= "\n";
    $script_str .= '    echo error running: $@';
    $script_str .= "\n  else\n";
    $script_str .= '    i_success=`expr $i_success + 1`';
    $script_str .= "\n  fi\n}\n\n";

    foreach my $gnome_schema_id (keys %$changed_settings)
    {
        my $settings = $changed_settings->{$gnome_schema_id};
        foreach my $gnome_key (keys %$settings)
        {
            my $data = $settings->{$gnome_key};
            my $cmd_str;
            if (defined($data))
            {
                $cmd_str = "$GSETTINGS set $gnome_schema_id $gnome_key \"$data\"";
            }
            else
            {
                next;
            }
            $script_str .= "$cmd_str\n";
            $script_str .= "verify_result $cmd_str\n";
        }
    }

    $script_str .= 'echo Successful update: $i_success';
    $script_str .= "\n";
    $script_str .= 'echo Failed update: $i_fail';
    $script_str .= "\n";
    $script_str .= 'exit $i_fail';
    $script_str .= "\n";


    # create script in temp dir
    my $tmpdir = GetTempDirPath(0);
    if (! defined($tmpdir))
    {
        ERROR_OUT("Cannot get temp dir path.");
        return undef;
    }

    my $file = "$tmpdir/$GSETTINGS_SCRIPT_FILE";
    my $rc = WriteFile($file, $script_str);
    if (! defined($rc))
    {
        ERROR_OUT("Fail to write file $file.");
        return undef;
    }

    $rc = chmod(0755, $file);
    if (! $rc)
    {
        ERROR_OUT("Cannot change $file mode to 0755");
        unlink($file);
        return undef;
    }

    DEBUG_OUT("Created script $file.");
    TRACE_OUT("Script content:\n$script_str");

    return $file;
}

#
# Update settings.
#
#   $_[0]:  script path
#   $_[1]:  user
#   $_[2]:  array reference of dbus session addresses
#
#   return: 1       - successful
#           0       - some setting not updated
#           undef   - failed
#
sub UpdateSettings($$$)
{
    my $script = $_[0];
    my $user = $_[1];
    my $dbus_session_addresses = $_[2];

    defined($script) or return undef;
    defined($user) or return undef;
    defined($dbus_session_addresses) or return undef;

    my $i_fail = 0;

    foreach my $addr (@$dbus_session_addresses)
    {
        my $cmd = [$SU, "-", $user, "-c", "DBUS_SESSION_BUS_ADDRESS=$addr XDG_RUNTIME_DIR= $script"];
        my ($rc, $output) = RunCmd($cmd);
        if (! defined($rc) or $rc ne '0')
        {
            $i_fail++;
            ERROR_OUT("Gnome 3 setting update script $script has $rc failures.");
            DEBUG_OUT("Command output:\n$output");
            next;
        }
    }

    return ($i_fail > 0) ? 0 : 1;
}

#
# Verify settings.
#
#   $_[0]:  user
#   $_[1]:  hash reference of changed settings
#
#   return: 1       - verified
#           0       - has error
#           undef   - failed
#
sub VerifySettings($$)
{
    my $user = $_[0];
    my $changed_settings = $_[1];

    defined($user) or return undef;
    defined($changed_settings) or return undef;

    my $ret = 1;

    my $sys = GetSysSettings($user);

    foreach my $gnome_schema_id (keys %$changed_settings)
    {
        my $settings = $changed_settings->{$gnome_schema_id};
        foreach my $gnome_key (keys %$settings)
        {
            my $data = $settings->{$gnome_key};
            my $sys_data = $sys->{$gnome_schema_id}{$gnome_key};
            my $correct = 1;
            if (defined($data))
            {
                if ($sys_data ne $data)
                {
                    $correct = 0;
                    $ret = 0;
                }
            }
            else
            {
                if (defined($sys_data))
                {
                    $correct = 0;
                    $ret = 0;
                }
            }

            if (! $correct)
            {
                DEBUG_OUT("  May not updated: $gnome_schema_id: $gnome_key  expected: [%s]  actual: [%s]", defined($data) ? $data : 'undefined', defined($sys_data) ? $sys_data : 'undefined');
            }
        }
    }

    return $ret;
}

#
# Decide whether need to do map, unmap, or do nothing.
#
#   $_[0]:  action
#
#   return: string      - map|unmap|skip
#           undef       - do nothing
#
sub GetAction($)
{
    my $action = $_[0];

    defined($action) or return undef;

    if ($action eq 'unmap')
    {
        return 'unmap';
    }

    my $reg_enable_gnome3 = CentrifyDC::GP::RegHelper->new($action, 'user', $REGKEY, 'EnableGnome3Settings', undef);
    $reg_enable_gnome3 or return undef;
    $reg_enable_gnome3->load();

    my $cur  = $reg_enable_gnome3->get('current');
    my $prev = $reg_enable_gnome3->get('previous');

    # if Gnome 3 GP is enabled, do map
    if (defined($cur) && $cur == 1)
    {
        DEBUG_OUT("Gnome 3 policy is enabled.");
        return 'map';
    }

    #
    # if Gnome 3 GP not configured or disabled, check previous setting.
    # if previous setting is enabled, then need to restore original setting,
    # so do unmap.
    #
    if (defined($prev) && $prev == 1)
    {
        DEBUG_OUT("Gnome 3 policy is turned off.");
        return 'unmap';
    }

    return 'skip';
}

#
# Create a dbus session for specified user.
#
# Get session address so that we can write settings to this session. Get pid
# so that the session can be killed later.
#
#   $_[0]:  user
#
#   return:
#      $1:  1       - successful
#           undef   - failed
#      $2:  dbus session pid
#      $3:  dbus session address
#
sub CreateDbusSession($)
{
    my $user = $_[0];

    defined($user) or return;

    my $pid;
    my $addr;

    my $cmd = [$SU, "-", $user, "-c", "$DBUSLAUNCH --sh-syntax"];

    my ($rc, $output) = RunCmd($cmd);
    if (! defined($rc) or $rc ne '0')
    {
        DEBUG_OUT("Fail to run command [$cmd].");
        return;
    }
    my @lines = split(/\n/, $output);

    foreach my $line (@lines)
    {
        #
        # DBUS_SESSION_BUS_ADDRESS='unix:abstract=/tmp/dbus-xxxxxxxxxx,guid=yyyyyyyyyyyyyyyyyyyyyyyyyyyyyyyy';
        # don't need the guid part
        #
        if ($line =~ m/^DBUS_SESSION_BUS_ADDRESS='(.+),guid=/)
        {
            $addr = $1;
            next;
        }
        # DBUS_SESSION_BUS_PID=xxxx;
        if ($line =~ m/^DBUS_SESSION_BUS_PID=([0-9]+)/)
        {
            $pid = $1;
            next;
        }
    }

    if (defined($addr) && defined($pid))
    {
        DEBUG_OUT("DBUS session address: $addr\nDBUS session pid: $pid");
        return (1, $pid, $addr);
    }
    else
    {
        ERROR_OUT("Fail to get DBUS session address or pid");
        return;
    }
}

#
# Kill the dbus session.
#
#   $_[0]:  pid
#
#   return: 1       - successful
#           0       - no need to kill
#           undef   - failed
#
sub KillDbusSession($)
{
    my $pid = $_[0];

    defined($pid) or return 0;

    DEBUG_OUT("Kill DBUS session $pid");
    my $ret = kill(SIGTERM, $pid);

    if ($ret)
    {
        DEBUG_OUT("DBUS session $pid killed successfully.");
        return 1;
    }
    else
    {
        ERROR_OUT("Failed to kill DBUS session $pid.");
        return undef;
    }
}


#
# Get all existing dbus sessions for specified user.
#
#   $_[0]:  user
#
#   return: array reference - dbus session addresses
#           undef           - failed
#
sub GetExistingDbusSessions($)
{
    my $user = $_[0];

    defined($user) or return undef;

    my $dbus_session_addresses = [];

    my $cmd = [$PS, "-u", $user];

    my ($rc, $output) = RunCmd($cmd);
    if (! defined($rc) or $rc ne '0')
    {
        DEBUG_OUT("Fail to run command [$cmd].");
        return undef;
    }

    my @lines = split(/\n/, $output);

    #
    # get all gnome-session of user, then get dbus session address from
    # each session's proc file.
    #
    foreach my $line (@lines)
    {
        # find gnome-session's pid
        if ($line =~ m/^ *([0-9]+).* gnome-session$/)
        {
            my $pid = $1;
            DEBUG_OUT("Found gnome session pid $pid");

            # get DBUS_SESSION_BUS_ADDRESS from proc file
            my $file = "/proc/$pid/environ";
            my $str = ReadFile($file);
            if (! defined($str))
            {
                ERROR_OUT("Cannot read $file");
                next;
            }

            my $addr;
            if ($str =~ m/DBUS_SESSION_BUS_ADDRESS=([^\0]+)/)
            {
                $addr = $1;
                # we don't need the guid part
                $addr =~ s/,guid=.*//;
                DEBUG_OUT("Found DBUS address: [$addr]");
                push(@$dbus_session_addresses, $addr);
            }
        }
    }

    return $dbus_session_addresses;
}

#
# Map all Gnome 3 settings.
#
#   $_[0]:  action (map|unmap) 
#   $_[1]:  username 
#
#   return: 1       - successful
#           0       - no need to change setting
#           undef   - failed
#
sub DoMap($$)
{
    my ($action, $user) = @_;

    defined($action) or return undef;
    defined($user) or return undef;

    my $b_create_dbus_session;

    # get relocatable schemas. will skip these schemas later.
    GetRelocatableSchemas($user);

    # dump all system settings
    my $sys_settings = GetSysSettings($user);
    if (! defined($sys_settings))
    {
        ERROR_OUT("Cannot get system settings.");
        return undef;
    }

    # should not happen, but if no setting, skip
    if (IsEmpty($sys_settings))
    {
        DEBUG_OUT("No system settings found.");
        return 0;
    }

    # get settings that exist in both registry and system
    my $settings = GetAllSettings($action, $user, $sys_settings);

    if (IsEmpty($settings))
    {
        DEBUG_OUT("No Gnome 3 settings need to be updated.");
        return 0;
    }

    # get settings that need to be changed.
    my $changed_settings = GetChangedSettings($settings, $action);
    if (IsEmpty($changed_settings))
    {
        DEBUG_OUT("No Gnome 3 settings changed.");
        return 0;
    }

    # get existing dbus session address
    my $dbus_session_addresses = GetExistingDbusSessions($user);
    if (! defined($dbus_session_addresses) or IsEmpty($dbus_session_addresses))
    {
        DEBUG_OUT("Cannot find existing DBUS session. Need to create one.");
        $b_create_dbus_session = 1;
    }
    else
    {
        #
        # if there are too many dbus sessions, then updating all sessions may
        # take too much time. in this case, just create a new session.
        #
        if (scalar(@$dbus_session_addresses) > $MAX_DBUS_SESSIONS)
        {
            DEBUG_OUT("Too many DBUS sessions, may take too long to update all sessions. Need to create one.");
            $b_create_dbus_session = 1;
        }
        else
        {
            $b_create_dbus_session = 0;
        }
    }

    my $rc;

    # create a dbus session for setting update
    my $pid;
    my $addr;

    if ($b_create_dbus_session)
    {
        ($rc, $pid, $addr) = CreateDbusSession($user);
        if (! defined($rc))
        {
            ERROR_OUT("Cannot create DBUS session.");
            return undef;
        }
        $dbus_session_addresses = [$addr];
    }

    # create script that use gsettings to update setting.
    my $script = CreateScript($changed_settings);
    if (! defined($script))
    {
        ERROR_OUT("Cannot create script to update setting.");
        # kill the dbus session
        if ($b_create_dbus_session)
        {
            $rc = KillDbusSession($pid);
            if (! defined($rc))
            {
                ERROR_OUT("Error occured when killing DBUS session pid $pid.");
                return undef;
            }
        }
    }

    #
    # update setting. should not return if there's any failure, as
    # need to kill the dbus session.
    #
    $rc = UpdateSettings($script, $user, $dbus_session_addresses);
    if (! defined($rc))
    {
        ERROR_OUT("Gnome 3 setting update failed.");
    }
    elsif ($rc == 0)
    {
        DEBUG_OUT("Some Gnome 3 settings not updated.");
    }
    else
    {
        DEBUG_OUT("All Gnome 3 settings updated correctly.");
    }

    if (! $rc)
    {
        VerifySettings($user, $changed_settings);
    }

    unlink($script);

    # kill the dbus session
    if ($b_create_dbus_session)
    {
        $rc = KillDbusSession($pid);
        if (! defined($rc))
        {
            ERROR_OUT("Error occured when killing DBUS session pid $pid.");
            return undef;
        }
    }

    return 1;
}



# >>> MAIN >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

my $args = CentrifyDC::GP::Args->new('user');

my $action = $args->action();
my $user = $args->user();

if (! IsGnome3())
{
    DEBUG_OUT("This mapper is for Gnome 3. Skip.");
    exit(0);
}

IsRoot() or FATAL_OUT("This script must run as root.");

CheckTools() or FATAL_OUT("Some required tools missing. Abort.");

CentrifyDC::GP::Registry::Load($user);

$action = GetAction($action);
defined($action) or FATAL_OUT("Cannot determine what action to take.");
if ($action eq 'skip')
{
    DEBUG_OUT("No need to change Gnome 3 settings.");
    exit(0);
}
DEBUG_OUT("Action: $action");

my $ret = DoMap($action, $user);

defined($ret) or FATAL_OUT();

