#!/bin/sh /usr/share/centrifydc/perl/run

##############################################################################
#
# Copyright (C) 2004-2020 Centrify Corporation. All rights reserved.
#
# Machine/user mapper script to import system NSSDB to specified application
# NSSDB.
#
# All commands are run as root.
#
#  Map:     import system NSSDB to specified application NSSDB
#
#  Unmap:   do nothing
#
# Parameters: <map|unmap> [username] mode
#   map|unmap   action to take
#   username    username
#   mode        mode (not used)
#
# Exit value:
#   0   Normal
#   1   Error
#   2   usage
#
##############################################################################

use strict;
use lib '/usr/share/centrifydc/perl';
use CentrifyDC::Include;
use CentrifyDC::GP::Args;
use CentrifyDC::GP::General qw(:debug RunCmd IsEmpty GetTempDirPath CreateTempFile CreateDir CreateFile ChangeOwner ReadFile);
use CentrifyDC::GP::CimsPath qw($CIMS_USR $CIMS_INSTALL);
use CentrifyDC::GP::RegHelper;
use POSIX qw(setuid setgid);


my $TEMP_DIR = GetTempDirPath(0);
defined($TEMP_DIR) or FATAL_OUT();


my $REGKEY = "software/policies/centrify/unixsettings/security/importnssdb";
my $sourcedir = "/etc/pki/nssdb";
my $openssl = "$CIMS_INSTALL/bin/openssl";
my $crldirectory = "/etc/pam_pkcs11/crls";
my $certdirectory = "/etc/pam_pkcs11/cacerts";

my $temp_der_form_crl_for_app = "$TEMP_DIR/cdctmpcrlforapp";

#takes in 1 argument which is new_app_location
sub crlimport
{
    my $new_app_location = $_[0];
    my $username = $_[1];
    my @user;
    my $crlfile;
    my $pid;
    opendir (DIR, $crldirectory) or die $!;
    # for all crls in /etc/pam_pkcs11/crls...
    while ($crlfile = readdir(DIR))
    {
        chomp($crlfile);
        if ($crlfile =~ m/^trust_.*\.crl$/)
        {
            # skip empty crl file
            if ( -z $crldirectory . "/" . $crlfile)
            {
                next;
            }

            my ($fh, $tmpfile) = CreateTempFile($temp_der_form_crl_for_app);
            if (! defined($fh) or ! defined($tmpfile))
            {
                die "Cannot create temp file";
            }
            close($fh);

            #crlutil doesn't accept PEM format for import so we make a temp DER format crl 
            `$openssl crl -in $crldirectory/$crlfile -inform PEM -out $tmpfile -outform DER`;

            defined ($pid = fork) or die "Cannot fork $!";
            if ($pid < 0)
            {
                die "Fork failed $!";
            }
            if ($pid == 0)
            {
                # child process
                @user = getpwnam($username);
                if (@user)
                {
                    # user no permission to read tmpfile created by openssl, need change ownership.
                    ChangeOwner($tmpfile, $username) or die "Cannot change $tmpfile owner to $username";

                    DEBUG_OUT("Setting to gid: [$user[3]]");
                    DEBUG_OUT("Setting to uid: [$user[2]]");
                    setgid $user[3] or die "Cannot switch group for $username $!";
                    setuid $user[2] or die "Cannot switch id to $username $!";
                }
                else
                {
                    die "Invalid user ID($username)";
                }

                my $new_app_profiles_ini = $new_app_location . "/profiles.ini"; 

                # import each crl into all profiles under the current user's application
                if (CreateDir($new_app_location) && CreateFile($new_app_profiles_ini))
                {
                    my $str = ReadFile($new_app_profiles_ini);
                    if (! defined($str))
                    {
                        DEBUG_OUT("Cannot read $new_app_profiles_ini");
                        next;
                    }
                    
                    my @lines = split(/\n/, $str);
                    
                    foreach my $line (@lines)
                    {
                        chomp $line;
                        if ($line =~ m/^Path=.*/)
                        {
                            $line =~ s/^Path=//;
                            DEBUG_OUT("Importing crl: [$crldirectory/$crlfile] to profile: [$new_app_location/$line]");
                            my ($ret, $out) = RunCmd(["/usr/bin/crlutil", "-I", "-i", $tmpfile, "-d", "$new_app_location/$line"]);
                            if ($ret != 1)
                            {
                                DEBUG_OUT("RunCmd failed: $ret,$out");
                            }
                        }
                    }
                }

                # kill the child process
                exit(1);
            }
            else
            {
                # parent process
                my $ret = waitpid($pid, 0);
                unlink($tmpfile);
                if ($ret < 0)
                {
                    die "waitpid failed for pid $pid";
                }
            }
        }
    }
    closedir(DIR);
}

sub certimport
{
    my $new_app_location = $_[0];
    my $username = $_[1];
    my @user;
    my $certfile;
    my $pid;
    defined ($pid = fork) or die "Cannot fork $!";
    if ($pid < 0)
    {
        die "Fork failed $!";
    }
    if ($pid == 0)
    {
        # child process
        @user = getpwnam($username);
        if (@user)
        {
            DEBUG_OUT("Setting to gid: [$user[3]]");
            DEBUG_OUT("Setting to uid: [$user[2]]");
            setgid $user[3] or die "Cannot switch group for $username $!";
            setuid $user[2] or die "Cannot switch id to $username $!";
        }
        else
        {
            die "Invalid user ID($username)";
        }

        my $new_app_profiles_ini = $new_app_location . "/profiles.ini"; 

        # create default profile
        if (index($new_app_location, "firefox") != -1 ||
            index($new_app_location, "thunderbird") != -1)
        {
            if (CreateDir($new_app_location))
            {
                CreateDir($new_app_location . "/" . $username . ".default");

                # generate an empty prefs.js
                my $ret = RunCmd(["/bin/touch", "$new_app_location/$username.default/prefs.js"]);
                if ($ret != 0)
                {
                    DEBUG_OUT("RunCmd failed: $ret");
                }

                # generate profiles.ini
                if (CreateFile($new_app_profiles_ini))
                {
                    my $profiles_ini_content = "";
                    $profiles_ini_content .= "[General]\n";
                    $profiles_ini_content .= "StartWithLastProfile=1\n";
                    $profiles_ini_content .= "\n";
                    $profiles_ini_content .= "[Profile0]\n";
                    $profiles_ini_content .= "Name=default\n";
                    $profiles_ini_content .= "IsRelative=1\n";
                    $profiles_ini_content .= "Path=$username.default\n";
                    $profiles_ini_content .= "\n";
                    open PROFILES_INI_OUTPUT, "> $new_app_profiles_ini";
                    print PROFILES_INI_OUTPUT $profiles_ini_content;
                    close PROFILES_INI_OUTPUT;
                }
            }
        }

        opendir (DIR, $certdirectory) or die $!;
        # for all certs in /etc/pam_pkcs11/cacerts...
        while ($certfile = readdir(DIR))
        {
            chomp($certfile);
            if ($certfile =~ m/^trust_.*\.cert$/)
            {
                # import each cert into all profiles under the current user's application
                if (CreateDir($new_app_location) && CreateFile($new_app_profiles_ini))
                {
                    my $str = ReadFile($new_app_profiles_ini);
                    if (! defined($str))
                    {
                        DEBUG_OUT("Cannot read $new_app_profiles_ini");
                        next;
                    }
                    
                    my @lines = split(/\n/, $str);
                    
                    foreach my $line (@lines)
                    {
                        chomp $line;
                        if ($line =~ m/^Path=.*/)
                        {
                            $line =~ s/^Path=//;
                            DEBUG_OUT("Importing cert: [$certdirectory/$certfile] to profile: [$new_app_location/$line]");
                            my $ret = RunCmd(["/usr/bin/certutil", "-A", "-t", "CT,C,C", "-i", "$certdirectory/$certfile", "-n", $certfile, "-d", "$new_app_location/$line"]);
                            if ($ret != 0)
                            {
                                DEBUG_OUT("RunCmd failed: $ret");
                            }
                        }
                    }
                }
            }
        }
        closedir(DIR);
        # kill the child process
        exit(1);
    }
    else
    {
        # parent process
        if (waitpid($pid, 0) < 0)
        {
            die "waitpid failed for pid $pid";
        }
    }
}

# >>> MAIN >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

my $args = CentrifyDC::GP::Args->new();

$args->isMap() or exit 0;

CentrifyDC::GP::Registry::Load($args->user());

my $reg_app_location = CentrifyDC::GP::RegHelper->new($args->action(), $args->class(), $REGKEY, undef, undef, 1);
$reg_app_location or FATAL_OUT("Cannot create RegHelper instance");
$reg_app_location->load('current');

my $err = 0;
my $user;
my $new_app_location;

my $app_locations = $reg_app_location->get('current');
if (! IsEmpty($app_locations))
{
    foreach my $app_location (@$app_locations)
    {
        DEBUG_OUT("Generic Application NSSDB location: [$app_location]");
        $user = $args->user();
        
        # removing the ~ from app_location and fitting 
        # in the fullpath of the user home dir

        $app_location =~ s/^\~\///;
        my $homedir = (getpwnam($user))[7];
        defined($homedir) or FATAL_OUT("Cannot get home dir of user [$user]");
        $new_app_location = "$homedir/$app_location";
        DEBUG_OUT("User specific Application NSSDB location: [$new_app_location]");

        # import the certificates
        certimport($new_app_location, $user);

        # import the crls
        crlimport($new_app_location, $user);
    }
}

$err and FATAL_OUT();

