#!/bin/sh /usr/share/centrifydc/perl/run

##############################################################################
#
# Copyright (C) 2014-2020 Centrify Corporation. All rights reserved.
#
# RHEL-specific machine mapper script to enable/disable smartcard screen lock
#
#  The script enable/disable smartcard screen lock.
#
#  Map:     enable or disable smartcard screen lock based on gp setting
#             Not configured:  disable
#             Enabled:         enable
#             Disabled:        disable
#
#  Unmap:   disable smartcard screen lock
#
# Parameters: <map|unmap> mode
#   map|unmap   action to take
#   mode        mode (not used)
#
# Exit value:
#   0   Normal
#   1   Error
#   2   usage
#
##############################################################################

use strict;
use lib '/usr/share/centrifydc/perl';
use CentrifyDC::Include;
use CentrifyDC::GP::Args;
use CentrifyDC::GP::Registry;
use CentrifyDC::GP::General qw(:debug RunCmd GetTempDirPath);
use CentrifyDC::GP::CimsPath qw($CIMS_USR $CIMS_INSTALL);
use CentrifyDC::GP::Mapper;
use CentrifyDC::GP::RegHelper;

my $REGKEY = 'software/policies/centrify/unixsettings/security/rhelsmartcardscreenlock';

# >>> SUB >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>


sub ConfigSmartCardLock($)
{
    my $action = $_[0];

    chomp($action);

    if ($action eq "1")
    {
        RunCmd(["$CIMS_INSTALL/sbin/rhel_authconfig.sh", "smartcardaction", "lock"]);
    }
    else
    {
        RunCmd(["$CIMS_INSTALL/sbin/rhel_authconfig.sh", "smartcardaction", "ignore"]);
    }
}

sub SmartCardRemovalAction($$)
{
    my $action = $_[0];
    my $user = $_[1];

    chomp($action);
    chomp($user);

    # execute launch_card_eventmgr.sh to serve screen lock.
    # NOTE: only execute it when login with smart card.

    my $TMP_DIR = GetTempDirPath(0);
    if (!defined($TMP_DIR))
    {
        FATAL_OUT('Cannot get temporary directory');
    }
    # smart card token tmp file
    my $SC_TMP_FILE    = $TMP_DIR . "/cdc-" . $user;
    # name mapping smart card token tmp file
    my $SC_NM_TMP_FILE = $TMP_DIR . "/CDC-NM-SC-TOKEN";

    if ( -e $SC_TMP_FILE or -e $SC_NM_TMP_FILE )
    {
        if ($action eq "1")
        {
            # launch in background cuz this script will wait for kded4 to fire up
            system("$CIMS_INSTALL/bin/launch_card_eventmgr.sh " . $user . " &");
        }
        else
        {
            RunCmd(["/usr/bin/killall", "card_eventmgr"]);
        }
    }
}

# >>> MAIN >>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>>

my $err = 0;
my $args = CentrifyDC::GP::Args->new();
my $user = $args->user();
my $class = $args->class();

CentrifyDC::GP::Registry::Load(undef);

my $reg = CentrifyDC::GP::RegHelper->new($args->action(), 'machine', $REGKEY, 'RHELSmartCardScreenLock', undef);
$reg or FATAL_OUT("Cannot create RegHelper instance");
$reg->load();

my $screenLock = $reg->get('current');

# Something other than "not configured"
if (defined($screenLock))
{
    # Mapping the gp
    if ($args->isMap())
    {
        if ($class eq "machine")
        {
            ConfigSmartCardLock($screenLock);
        }
        elsif ($class eq "user")
        {
            SmartCardRemovalAction($screenLock, $user);
        }
    }
    # Unmapping the gp
    else
    {
        if ($class eq "machine")
        {
            ConfigSmartCardLock("0");
        }
        elsif ($class eq "user")
        {
            SmartCardRemovalAction("0", $user);
        }
    }
}

$err and FATAL_OUT();
