#!/bin/sh /usr/share/centrifydc/perl/run.cda

##############################################################################
#
# Copyright (C) 2014-2020 Centrify Corporation. All rights reserved.
#
# This script is used to modify gdm configure files
#
# arguments:
#       please see sub usage
# return:
#       0 if succeed
#       1 if meet error
#
##############################################################################

use warnings;
use strict;
use lib "/usr/share/centrifydc/perl";
use lib "/usr/share/centrifydc/scripts/da";
use CentrifyDA::Include;
use Getopt::Long qw(GetOptions);
use autoedit_constants qw($CDA_Modified_Comment $CDA_Version);
use autoedit_file_lock;
use CentrifyDA::General qw(:debug :system);
use Fcntl qw(:mode);

use constant DEFAULT_CFG_FILE => "/etc/gdm/PreSession/Default";
use constant DAX_STARTUP_SCRIPT => "/usr/sbin/rundax.sh";
my $cfg_file = DEFAULT_CFG_FILE;
my $is_gdm_installed;
my $is_gdm_hooked;
my $do_revert;

my $quote_dax_startup_script = quotemeta(DAX_STARTUP_SCRIPT);

sub usage
{
    warn << "END";
    This script is used to modify gdm configure file
    usage:
    --cfg_file          the full path of gdm configure file
    --is_gdm_installed  check if desktop enviroment is installed
    --is_gdm_hooked     check if the start-up script of dax is found in gdm configure file
    --do_revert         set it if want to revert our changes
END
    exit 1;
}

#
# Check if the file is secure. The target file is secure if
#     1. The file is owned by root
#     2. The file is not world writable
#     3. The file is not symlink
#     4. It has no more than one hard link for the file/inode
#
# $_[0]: gdm configure file
#
# Return value:
#       0 - Means that it's not secure
#       1 - Means that it's secure
#
sub isFileSecure
{
    my ($cfgName) = @_;

    if (! -e $cfgName)
    {
        DEBUG_OUT("$cfgName does not exist.");
        return 0;
    }

    my ($mode, $nlink, $uid) = (stat($cfgName))[2, 3, 4];

    if (!S_ISREG($mode))
    {
        DEBUG_OUT("$cfgName is not a regular file.");
        return 0;
    }

    my $l_mode = (lstat($cfgName))[2];
    if (S_ISLNK($l_mode))
    {
        DEBUG_OUT("$cfgName is a symlink file.");
        return 0;
    }

    if ($uid != 0)
    {
        DEBUG_OUT("$cfgName is not owned by root.");
        return 0;
    }

    if ($mode & S_IWOTH)
    {
        DEBUG_OUT("$cfgName is world writable.");
        return 0;
    }

    if ($nlink != 1)
    {
        DEBUG_OUT("$cfgName has more than one hard link.");
        return 0;
    }

    return 1;
}

#
# Process gdm configure file
#
# $_[0]:     the full path of gdm configure file
# $_[1]:     do revert
#
# Return value:
#       0 - If meet error
#       1 - If succeed
#
sub processConfFile
{
    my ($cfgName, $revert) = @_;

    my $flock = autoedit_file_lock::new("$cfgName");
    if (!defined $flock)
    {
        warn "cannot lock file $cfgName";
        return 1;
    }

    $cfgName = TraverseSymLink($cfgName);
    if (!defined $cfgName)
    {
        DEBUG_OUT("meet error when check symlink $cfgName");
        return 1;
    }

    my $cfgLines = ReadFile("$cfgName");
    if (!defined $cfgLines)
    {
        DEBUG_OUT("cannot read $cfgName");
        return 1;
    }

    my $newLines = << "END";
$CDA_Modified_Comment $CDA_Version
@{[DAX_STARTUP_SCRIPT]} &
END

    if ($revert)
    {
        # remove CentrifyDA additions
        $cfgLines =~ s/ *#.*centrifyda.*\n//img;
        $cfgLines =~ s/$quote_dax_startup_script.*\n//img;
    }
    elsif ($cfgLines =~ m/$quote_dax_startup_script/)
    {
        return 0;
    }
    else
    {
        if ($cfgLines !~ /\n$/)
        {
            # not end with "\n" ?
            $cfgLines .= "\n";
        }
        $cfgLines .= $newLines;
    }

    my $backup = "$cfgName.pre_cda";
    if (!$revert)
    {
        if (!CopyFile($cfgName, $backup))
        {
            ERROR_OUT("Cannot backup $cfgName");
            return 1;
        }
    }
    else
    {
        # install.sh and rpm/pkg scripts may override config files with
        # the backup files if found them exist.
        # To avoid affecting these code logic, we'd better delete 
        # backup file after "revert"
        if (-e $backup)
        {
            unlink $backup;
        }
    }

    if (!WriteFile($cfgName, $cfgLines))
    {
        ERROR_OUT("Cannot write $cfgName");
        return 1;
    }

    return 0;
}

#
# Check if Desktop enviroment (gdm) is installed
#
# $_[0]: gdm configure file
#
# Return value:
#       0 - Means that it's installed
#       1 - Means that it's not installed
#
sub isGdmInstalled
{
    my ($cfgName) = @_;
    if (-e $cfgName)
    {
        return 0;
    }
    
    return 1;
}

#
# Check if the start-up script of dax is found in gdm configure file
#
# $_[0]: gdm configure file
#
# Return value:
#       0 - Means that it's found
#       1 - Means that it's not found
#
sub isGdmHooked
{
    my ($cfgName) = @_;
    my $flock = autoedit_file_lock::new("$cfgName");
    if (!defined $flock)
    {
        warn "cannot lock file $cfgName";
        return 1;
    }

    $cfgName = TraverseSymLink($cfgName);
    if (!defined $cfgName)
    {
        DEBUG_OUT("meet error when check symlink $cfgName");
        return 1;
    }

    my $cfgLines = ReadFile("$cfgName");
    if (!defined $cfgLines)
    {
        DEBUG_OUT("cannot read $cfgName");
        return 1;
    }
    
    if ($cfgLines =~ m/$quote_dax_startup_script/)
    {
        return 0;
    }
    
    return 1;
}


GetOptions( "cfg_file=s" => \$cfg_file,
            "is_gdm_installed" => \$is_gdm_installed,
            "is_gdm_hooked" => \$is_gdm_hooked,
            "do_revert" => \$do_revert,
            ) or usage;

# Check if Desktop enviroment (gdm) is installed
if ($is_gdm_installed)
{
    my $ret = isGdmInstalled($cfg_file);
    exit $ret;
}

# Check if the start-up script of dax is found in gdm configure file
if ($is_gdm_hooked)
{
    my $ret = isGdmHooked($cfg_file);
    exit $ret;
}

# Make sure that gdm configure file is secure before processing it.
if (!isFileSecure($cfg_file))
{
    ERROR_OUT("$cfg_file is not secure.");
    exit 1;
}

# Update gdm conf file
if (-e $cfg_file)
{
    my $ret = processConfFile($cfg_file, $do_revert);
    exit $ret;
}

exit 0;
